###########################################
#######  CALCULATE DIFF AND DIST MATRIX ###
###########################################

require(ncdf4)
require(lubridate)
require(raster)
require(geosphere)

# 1. mean monthly difference between TabsD and Measurements 
# 2. distance matrix

### this is done for all temperature measurements if set here to "all_temp". Needed stations are then selected based on ID.
scen = "all_temp"
tdiff = T
tdist = F
detr = T

## read TabsD  (TabsD offset)
args <- commandArgs(TRUE)
if (length(args)>0) {
  scen = args[1]
}

if(detr){
  print("load detrenden")
  tabs <- "data_input/TabsD_ch01r.swiss.lv95_detrended_19610101-20201231.nc"
  tabs_clim <- "data_input/TabsD_ch01r.swiss.lv95_detrended_clim_19610101-20201231.nc"
  tabs_deseas <- "data_input/TabsD_ch01r.swiss.lv95_detrended_deseas_19610101-20201231.nc"
} else {
  print("load regular")
  tabs <- "data_input/TabsD_ch01r.swiss.lv95_19610101-20201231.nc"
  tabs_clim <- "data_input/TabsD_ch01r.swiss.lv95_clim_19610101-20201231.nc"
  tabs_deseas <- "data_input/TabsD_ch01r.swiss.lv95_deseas_19610101-20201231.nc"
}


### read inventory
source("C011_scenarios.R")
station_coords <- XLConnect::readWorksheetFromFile("data_input/arm_stations_inventory_withnbcn.xlsx",sheet=1, startRow=2)[,c("ID","lat_ref","lon_ref","E_lv95_ref","N_lv95_ref","alt_ref")]
temp_stns <- scenarios[[scen]][grepl("_ta",scenarios[["all_temp"]])]
station_coords <- station_coords[which(station_coords$ID%in%temp_stns),]
station_coords$E_lv95_ref <- as.numeric(station_coords$E_lv95_ref)
station_coords$N_lv95_ref <- as.numeric(station_coords$N_lv95_ref)
station_coords$lon_ref <- as.numeric(station_coords$lon_ref)
station_coords$lat_ref <- as.numeric(station_coords$lat_ref)
station_coords$alt_ref <- as.numeric(station_coords$alt_ref)


#################################################################
### 1. mean monthly difference between TabsD and Measurements ###
#################################################################

if(tdiff){
  print("calc diff")
  
  load(paste0("data_input/analogue_stationdata",switch(detr+1,"_","_detrended_refperiod_"),scen,".RData"))
  ta_ind <- which(grepl("_ta",colnames(TOT)))
  
  nc <- nc_open(tabs)
  lons <- ncvar_get(nc,"lon")
  lats <- ncvar_get(nc,"lat")
  time <- as.Date(ncvar_get(nc,"time"), origin="1900-01-01")
  coords <- data.frame(lon=as.vector(lons), lat=as.vector(lats))
  
  df_ta_diff <- data.frame(matrix(NA, ncol=15,nrow=length(ta_ind)))
  df_ta_diff[,1] <- colnames(TOT)[ta_ind]
  
  t2 <- 0
  for (t in ta_ind){
    
    print(colnames(TOT)[t])
    t2 <- t2+1
    df_ta_diff[t2,2] <- station_coords[which(station_coords$ID==colnames(TOT)[t]),"E_lv95_ref"]
    df_ta_diff[t2,3] <- station_coords[which(station_coords$ID==colnames(TOT)[t]),"N_lv95_ref"]
    
    ## get coordinates of station
    xlon <- station_coords[which(station_coords$ID==colnames(TOT)[t]),"lon_ref"]
    xlat <- station_coords[which(station_coords$ID==colnames(TOT)[t]),"lat_ref"]
    
    options(digits=14)
    
    i_grid <- which.min(distGeo(c(xlon,xlat), coords))
    i <- ifelse(colnames(TOT)[t] == "ZUG_ta", 2, 1)
    i_lonlat <- which(lons == coords$lon[i_grid], arr.ind = T)[i,]
    gc <- ncvar_get(nc, "TabsD", c(i_lonlat[1],i_lonlat[2],1), c(1,1,-1))
  
    stn <- TOT[TOT$date%in%time,c(1,t)]
    diff <- stn[,2] - gc
    df_ta_diff[t2,4:ncol(df_ta_diff)] <- aggregate(diff,list(month(stn$date)), mean, na.rm=T)$x
  
    }
  
  colnames(df_ta_diff) <- c("ID", "lon","lat",1:12)
  write.table(df_ta_diff, file=paste0("data_input/diff_station_raster_all_temp.txt"), row.names = F, col.names = T)
  
}


#################################################################
### 2. distance matrix                                       ###
#################################################################

if(tdist){
  stn_coords <- cbind(as.numeric(unlist(station_coords[,"lon_ref"])), as.numeric(unlist(station_coords[,"lat_ref"])))
  CHalt <- raster("data_input/topo.swiss02_ch01r.swisscors.nc")
  ## convert to wgs84?
  crs(CHalt) <- CRS("+init=epsg:21781")
  print("calculate distance matrix")
  CHalt_vals <- data.frame(rasterToPoints(CHalt))
  colnames(CHalt_vals) <- c("lon","lat","alt")
  coordinates(CHalt_vals) <- c("lon", "lat")
  proj4string(CHalt_vals) <- CRS("+init=epsg:21781") # WGS 84
  CRSnew <- CRS("+init=epsg:4326")
  CHalt_wgs <- spTransform(CHalt_vals, CRSnew)
  CHalt_wgs_vals <- data.frame(lon=round(as.numeric(coordinates(CHalt_wgs)[,1]),5),lat=round(as.numeric(coordinates(CHalt_wgs)[,2]),5),alt=CHalt_vals$alt)
  
  ## add grid cells for the outside stations
  station_coords$ID%in%c("MIL_ta","TOR_ta","HOH_ta","ROV_ta")
  outside_vals <- station_coords[match(c("MIL_ta","TOR_ta","HOH_ta","ROV_ta"),station_coords$ID),c("lon_ref","lat_ref","alt_ref")]
  colnames(outside_vals) <- c("lon","lat","alt")
  CHalt_wgs_vals <- rbind(outside_vals,CHalt_wgs_vals)
  
  mat1 <- matrix(rep(as.numeric(CHalt_wgs_vals$alt), each=length(station_coords$ID)),nrow= length(station_coords$ID),ncol=length(CHalt_wgs_vals$alt))
  stat <- get_cellnum(CHalt_wgs_vals[,1:2], stn_coords)
  names(stat)  <- station_coords$ID
  mat2 <- matrix(rep(as.numeric(CHalt_wgs_vals$alt[stat]), each=length(CHalt_wgs_vals$alt)),nrow= length(stat),ncol=length(CHalt_wgs_vals$alt), byrow=T)
  alt_mat <- round(mat2-mat1)
  rownames(alt_mat) <- station_coords$ID
  save(alt_mat, file=paste0("data_input/","alt_mat_all_temp.RData"))
  rm(CHalt, mat1, mat2, stat)
  
  ### Create Distance matrix for inraster
  dist_mat <- t(mcsapply(1:nrow(stn_coords),function(x) {geosphere::distGeo(p1=CHalt_wgs_vals[,1:2],p2=stn_coords[x,])/1000}, mc.cores=round(0.5*parallel::detectCores())))
  rownames(dist_mat) <- station_coords$ID
  save(dist_mat, file=paste0("data_input/","dist_mat_all_temp.RData"))
}
