#!/anaconda3/envs/my_env/bin/python

# -----------------------------------------------
# validation of precipitation
# 29.06.2022
# -----------------------------------------------

import numpy as np
import pandas as pd
import xarray as xr
import bottleneck
from dask.distributed import Client

import sys
import glob
import time
import warnings  # to omit warnings for spatial correlation (as division by zero (no rain days) is not possible)
import os
import dask

client = Client(n_workers=2, threads_per_worker=4, memory_limit='20GB')
client
    
# -----------------------------------------------
# define functions
# -----------------------------------------------

rmse  = lambda x,y: np.sqrt(np.mean((x - y)**2, axis = -1))
mae   = lambda x,y: np.mean(np.abs(x-y), axis = -1)
bias  = lambda x,y: y.mean(axis = -1) - x.mean(axis = -1)
brier = lambda x,y: np.mean(((y > brier_threshold) * 1 - (x > brier_threshold) * 1 ) ** 2, axis = -1)

def covariance_gufunc(x, y):
    return ((x - x.mean(axis=-1, keepdims=True))
            * (y - y.mean(axis=-1, keepdims=True))).mean(axis=-1)

def pearson_correlation_gufunc(x, y):
    return covariance_gufunc(x, y) / (x.std(axis=-1) * y.std(axis=-1))

def spearman_correlation_gufunc(x, y):
    x_ranks = bottleneck.rankdata(x, axis=-1)
    y_ranks = bottleneck.rankdata(y, axis=-1)
    return pearson_correlation_gufunc(x_ranks, y_ranks)

## define wrapper functions for apply_ufunc

def spearman_correlation(x, y, dim):
    return xr.apply_ufunc(
        spearman_correlation_gufunc, x, y,
        input_core_dims=[[dim], [dim]],
        dask='parallelized',
        output_dtypes=[float])

def pearson_correlation(x, y, dim):
    return xr.apply_ufunc(
        pearson_correlation_gufunc, x, y,
        input_core_dims=[[dim], [dim]],
        dask='parallelized',
        output_dtypes=[float])

def bias_calculation(x, y, dim):
    return xr.apply_ufunc(
        bias, x, y,
        input_core_dims=[[dim], [dim]],
        dask='parallelized',
        output_dtypes=[float])

def rmse_calculation(x, y, dim):
    return xr.apply_ufunc(
        rmse, x, y,
        input_core_dims=[[dim], [dim]],
        dask='parallelized',
        output_dtypes=[float])

def mae_calculation(x, y, dim):
    return xr.apply_ufunc(
        mae, x, y,
        input_core_dims=[[dim], [dim]],
        dask='parallelized',
        output_dtypes=[float])

def brier_calculation(x, y, dim):
    return xr.apply_ufunc(
        brier, x, y,
        input_core_dims=[[dim], [dim]],
        dask='parallelized',
        output_dtypes=[float])
        
        
### the script is called from the console, otherwise uncomment.
sys.argv = ["", "/scratch3/noemi/wear/swiss_arm/code/output/ARM_run_2022-09-06/scen3/Qmap_2022-09-14/"]

print("start")

# omit warnings (see above)
warnings.filterwarnings("ignore")

# get start time
stm = time.time()

# define threshold for Brier score: rain/no rain (in mm)
brier_threshold = 1

# check if file is valid
if len(sys.argv) < 2:
    sys.exit("indicate input file for validation")

if not os.path.isdir(str(sys.argv[1])):
    sys.exit("file path invalid")

# create validation output folder
indir = os.path.dirname(str(sys.argv[1]))
exp = indir.split('/')
outdir = indir + "/validation_" + time.strftime("%Y-%m-%d") + "/"
print(outdir)
if not os.path.isdir(outdir):
    try:
        os.mkdir(outdir)
    except OSError:
        print("")
    else:
        print("")

# --------------------------------------------------------------------------------------------------------------------
# READ DATA
# --------------------------------------------------------------------------------------------------------------------
print("read observation data")

obs_ifolder = "/scratch3/noemi/wear/swiss_arm/code/data_input/"
obs_ifile = "RhiresD_ch01h.swiss.lv95_1961-01-01-2020-12-31_masked.nc"
precip_obs = xr.open_dataset(obs_ifolder + obs_ifile, chunks = dict(E = 10))
precip_obs = precip_obs["RhiresD"]  # precip_obs to dataarray (not dataset)
precip_obs

rec_ifile = str(sys.argv[1])
allfiles = glob.glob(f'{rec_ifile}*1961-01-01-2020-12-31.nc')
allfiles

print("read reconstruction data")
# this if for reading several files independent
#precip_rec = xr.open_mfdataset(paths = allfiles, engine = "netcdf4", chunks = dict(E = 10)).chunk({'time':-1})
precip_rec = xr.open_dataset(allfiles[0], chunks = dict(E = 10))
precip_rec = precip_rec["precip"]
precip_rec = precip_rec.sortby(precip_rec.N)
precip_rec

# --------------------------------------------------------------------------------------------------------------------
# Start validation
# --------------------------------------------------------------------------------------------------------------------


seasons = {"MAM": np.arange(3,6),"JJA":np.arange(6,9),"SON":np.arange(9,12),"DJF": [12,1,2],"annual":np.arange(1,13)}
validation = []


for name,value in seasons.items():
    print(value)
    print(name)
    
    obs_seas = precip_obs.sel(time=np.isin(precip_obs.time.dt.month,value))
    rec_seas = precip_rec.sel(time=np.isin(precip_rec.time.dt.month,value))
    
    print("spcorr")
    spcorr_out = spearman_correlation(obs_seas, rec_seas, "time").compute()
    print("bias")
    bias_out = bias_calculation(obs_seas, rec_seas, "time").compute()
    print("rmse")
    rmse_out = rmse_calculation(obs_seas, rec_seas, "time").compute()
    print("mae")
    mae_out = mae_calculation(obs_seas, rec_seas, "time").compute()
    print("brier")
    brier_out = brier_calculation(obs_seas, rec_seas, "time").compute()
    
    metrics = xr.concat([spcorr_out, bias_out, rmse_out, mae_out, brier_out], dim ="metrics", fill_value = np.nan)
    newcoords = ["corr_" + name, "bias_"  + name,"rmse_"  + name, 
                 "mae_"  + name, "brier_"  + name]    
    metrics = metrics.assign_coords(metrics=newcoords)
    
    metrics.E.attrs['axis'] = "X"
    metrics.E.attrs['units'] = "meters_east"
    metrics.E.attrs['long_name'] = "swiss_easting"
    metrics.N.attrs['axis'] = "Y"
    metrics.N.attrs['units'] = "meters_north"
    metrics.E.attrs['long_name'] = "swiss_northing"
    
    validation.append(metrics)
    
# --------------------------------------------------------------------------------------------------------------------
# Write file
# --------------------------------------------------------------------------------------------------------------------
ds = xr.concat([validation[0],validation[1],validation[2],validation[3],validation[4]], dim="metrics")
ds = ds.rename("vals")
ds.to_netcdf(outdir + time.strftime("%Y-%m-%d") + '_validation_maps_precip.nc')    
    
print("end")    
